"""Work with WCS objects for TESS"""

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS as astropyWCS
from astropy.wcs import Sip

from . import pointings, log  # noqa: E402
from . import pixel_corners, rcolumns, rrows
from .utils import (
    _load_support_dicts,
    _load_warp_matrices,
    _load_wcs_database,
    angle_to_matrix,
    deprecated,
    get_M,
)


class WCS(astropyWCS):
    """A special subclass of astropy.wcs.WCS

    This class allows us to add attributes to the WCS header and add class methods
    """

    wcs_dicts = _load_wcs_database()

    (
        xs,
        ys,
        xcent,
        ycent,
        M,
        sip_dict,
    ) = _load_support_dicts()

    Ms, offset_weights = _load_warp_matrices()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    @deprecated("from_sector")
    def from_archive(cls, sector: int, camera: int, ccd: int):
        """Load a WCS from archival TESS data"""
        return cls.from_sector(sector=sector, camera=camera, ccd=ccd)

    @classmethod
    def from_sector(cls, sector: int, camera: int, ccd: int):
        """Load a WCS from archival TESS data"""
        if sector in list(cls.wcs_dicts.keys()):
            wcs = cls(naxis=2)
            wcs.sector, wcs.camera, wcs.ccd = sector, camera, ccd
            wcs.ra = cls.wcs_dicts[sector]["ra"]
            wcs.dec = cls.wcs_dicts[sector]["dec"]
            wcs.roll = cls.wcs_dicts[sector]["roll"]
            wcs.pixel_shape = (rrows, rcolumns)
            wcs.wcs.ctype = ["RA---TAN-SIP", "DEC--TAN-SIP"]
            wcs.wcs.cunit = ["deg", "deg"]
            wcs.wcs.radesys = "ICRS"
            wcs.wcs.crpix = cls.wcs_dicts[sector][camera][ccd]["crpix0"]
            wcs.wcs.crval = cls.wcs_dicts[sector][camera][ccd]["crval0"]
            wcs.wcs.cdelt = [1, 1]
            wcs.wcs.pc = cls.wcs_dicts[sector][camera][ccd]["cd"]
            wcs.sip = Sip(
                *[cls.sip_dict[attr][camera][ccd] for attr in ["a", "b", "ap", "bp"]],
                [1045, 1001],
            )
            return wcs
        else:
            log.warning(
                f"Data for Sector {sector} has not been archived yet. This function will return the predicted WCS for Sector {sector}."
            )
            ra, dec, roll = [
                i
                for i in pointings[pointings["Sector"] == 3][["RA", "Dec", "Roll"]][
                    0
                ].values()
            ]
            wcs = cls.predict(
                ra=ra, dec=dec, roll=roll, camera=camera, ccd=ccd, warp=True
            )
            wcs.sector = sector
            return wcs

    @classmethod
    def predict(
        cls,
        ra: float,
        dec: float,
        roll: float,
        camera: int,
        ccd: int,
        warp: bool = True,
    ):
        """Predict a WCS based on RA, Dec and roll


        warp : bool
            Whether to warp by the best fit TESS Camera/CCD warp. This is recommended,
            otherwise solution will be off by up to 10 pixels. This must be switched off
            if fitting for the warp.
        """
        if (ccd == 1) | (ccd == 3):
            C = np.asarray([[1, -1], [-1, 1]])
        elif (ccd == 2) | (ccd == 4):
            C = np.asarray([[1, 1], [1, 1]])

        c_boresight = SkyCoord(ra, dec, unit="deg")
        R = angle_to_matrix(-roll)

        # initialize object
        wcs = cls(naxis=2)
        wcs.ra, wcs.dec, wcs.roll, wcs.camera, wcs.ccd = ra, dec, roll, camera, ccd
        wcs.pixel_shape = (rrows, rcolumns)

        # center of the CCD
        xc, yc = cls.xcent[camera][ccd], cls.ycent[camera][ccd]
        r, phi = np.hypot(xc, yc) * u.deg, np.arctan2(yc, xc) * u.rad
        c_ccd = c_boresight.directional_offset_by(
            position_angle=phi - roll * u.deg, separation=r
        )

        wcs.wcs.ctype = ["RA---TAN-SIP", "DEC--TAN-SIP"]
        wcs.wcs.cunit = ["deg", "deg"]
        wcs.wcs.radesys = "ICRS"
        #        wcs.wcs.crpix = np.asarray([rrows / 2, rcolumns / 2])
        wcs.wcs.crpix = np.asarray([1045, 1001])
        wcs.wcs.cdelt = [-1, 1]
        # center is the ccd center in world space
        wcs.wcs.crval = [c_ccd.ra.deg, c_ccd.dec.deg]
        wcs.sip = Sip(
            *[cls.sip_dict[attr][camera][ccd] for attr in ["a", "b", "ap", "bp"]],
            [1045, 1001],
        )
        # rotate the corners of the CCD to the correct roll
        x, y = R.dot(np.vstack([cls.xs[camera][ccd], cls.ys[camera][ccd]]))
        r, phi = np.hypot(x, y) * u.deg, np.arctan2(y, x) * u.rad

        # positions of the corners in world space
        c = c_boresight.directional_offset_by(position_angle=phi, separation=r)
        r, phi = c_ccd.separation(c), c_ccd.position_angle(c)

        # position of the corners in detector space
        truth = np.asarray([r * np.cos(phi), r * np.sin(phi)])

        # Our estimate of where the corners should in world space
        c = wcs.pixel_to_world(*pixel_corners.T)
        r, phi = c_ccd.separation(c), c_ccd.position_angle(c)
        # Our estimate of where the corners should in detector space
        approx = np.asarray([r * np.cos(phi), r * np.sin(phi)])

        # fit a transformation matrix
        matrix = get_M(truth.T, approx.T)[:2, :2]
        wcs.wcs.pc = matrix.dot(angle_to_matrix(180)).dot(
            np.asarray([[0.0593726, 0.00123252], [0.00123252, 0.0593726]]) * (C)
        )
        if warp:
            wcs.wcs.pc = wcs.wcs.pc.dot(cls.Ms[camera][ccd])
            ecl_roll = wcs.ecl_roll.rad
            X = np.asarray([np.sin(ecl_roll), np.cos(ecl_roll), 1])
            wcs.wcs.crpix += X.dot(cls.offset_weights[camera][ccd])
        return wcs

    def to_header(self, key=None):
        hdr = super().to_header(relax=True, key=key)
        self._update_header(hdr)
        return hdr

    def to_fits(self, **kwargs):
        hdu = fits.PrimaryHDU(header=self.to_header())
        hdulist = fits.HDUList([hdu])
        return hdulist

    def _update_header(self, hdr):
        hdr["AUTHOR"] = "tesswcs"
        for attr in ["ra", "dec", "roll", "sector", "camera", "ccd"]:
            if hasattr(self, attr):
                hdr[attr] = getattr(self, attr)
        return hdr

    @property
    def ecl_roll(self):
        """returns the roll angle of the pointing with respect to the ecliptic plane.
        This is handy for figuring out which are northern hemisphere pointings, and helping with warping.
        """
        c = SkyCoord(self.ra, self.dec, unit="deg")
        ecl = c.transform_to("geocentricmeanecliptic")

        # The roll angle w.r.t the ecliptic plane
        ecl_roll = ecl.position_angle(
            c.directional_offset_by(-self.roll * u.deg, 1 * u.arcsecond).transform_to(
                "geocentricmeanecliptic"
            )
        )
        return ecl_roll



"""Utilities for getting WCS"""

import bz2
import functools
import json
import os
import warnings

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.coordinates import Angle, SkyCoord

from . import PACKAGEDIR, log


def deprecated(replacement=None):
    """
    A decorator to mark functions as deprecated.

    Args:
        replacement (str): The name of the replacement function to include in the warning.
    """

    def decorator(func):
        replacement_msg = f" Use '{replacement}' instead." if replacement else ""
        message = f"The function '{func.__name__}' is deprecated.{replacement_msg}"

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                message,
                DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _fix_keys(dict):
    return {
        int(key): np.asarray(item) if isinstance(item, list) else item
        for key, item in dict.items()
    }


def _load_wcs_database():
    log.debug("Loading WCS data from file")
    filename = f"{PACKAGEDIR}/data/TESS_wcs_data.json.bz2"
    if not os.path.isfile(filename):
        raise ValueError(
            "`tesswcs` database doesn't exist, run `tesswcs.database.update_database()`."
        )
    with bz2.open(filename, "rt", encoding="utf-8") as f:
        wcs_dict = json.load(f)
    log.debug("Adding correct units to WCS data")
    _wcs_dicts = {}
    for sector in wcs_dict.keys():
        _wcs_dicts[int(sector)] = {
            "ra": wcs_dict[sector]["ra"],
            "dec": wcs_dict[sector]["dec"],
            "roll": wcs_dict[sector]["roll"],
        }
        for camera in "1234"[::-1]:
            _wcs_dicts[int(sector)][int(camera)] = {}
            for ccd in "1234"[::-1]:
                _wcs_dicts[int(sector)][int(camera)][int(ccd)] = {
                    "pa": Angle(wcs_dict[sector][camera][ccd]["pa"] * u.rad),
                    "sep": Angle(wcs_dict[sector][camera][ccd]["sep"] * u.deg),
                    "pa_sip": Angle(wcs_dict[sector][camera][ccd]["pa_sip"] * u.rad),
                    "sep_sip": Angle(wcs_dict[sector][camera][ccd]["sep_sip"] * u.deg),
                    "ccd_center_crval": wcs_dict[sector][camera][ccd][
                        "ccd_center_crval"
                    ],
                    "ccd_center_crval_sip": wcs_dict[sector][camera][ccd][
                        "ccd_center_crval_sip"
                    ],
                    "ccd_center_crpix": wcs_dict[sector][camera][ccd][
                        "ccd_center_crpix"
                    ],
                    "crval0": wcs_dict[sector][camera][ccd]["crval0"],
                    "crpix0": wcs_dict[sector][camera][ccd]["crpix0"],
                    "cd": np.asarray(wcs_dict[sector][camera][ccd]["cd"]),
                    "sip_a_order": wcs_dict[sector][camera][ccd]["sip_a_order"],
                    "sip_a": np.asarray(wcs_dict[sector][camera][ccd]["sip_a"]),
                    "sip_b": np.asarray(wcs_dict[sector][camera][ccd]["sip_b"]),
                    "sip_ap": np.asarray(wcs_dict[sector][camera][ccd]["sip_ap"]),
                    "sip_bp": np.asarray(wcs_dict[sector][camera][ccd]["sip_bp"]),
                    "corner_pa": Angle(
                        wcs_dict[sector][camera][ccd]["corner_pa"] * u.rad
                    ),
                    "corner_sep": Angle(
                        wcs_dict[sector][camera][ccd]["corner_sep"] * u.deg
                    ),
                    "corner": SkyCoord(
                        wcs_dict[sector][camera][ccd]["corner_ra"],
                        wcs_dict[sector][camera][ccd]["corner_dec"],
                        unit="deg",
                    ),
                    "corner_sip_pa": Angle(
                        wcs_dict[sector][camera][ccd]["corner_pa"] * u.rad
                    ),
                    "corner_sip_sep": Angle(
                        wcs_dict[sector][camera][ccd]["corner_sep"] * u.deg
                    ),
                    "corner_sip": SkyCoord(
                        wcs_dict[sector][camera][ccd]["corner_sip_ra"],
                        wcs_dict[sector][camera][ccd]["corner_sip_dec"],
                        unit="deg",
                    ),
                }
    return _wcs_dicts


def _load_support_dicts():
    log.debug("Loading support dictionaries from file")

    support_dicts = []

    for var in ["xs", "ys", "xcent", "ycent", "M"]:
        filename = f"{PACKAGEDIR}/data/TESS_wcs_{var}.json.bz2"
        with bz2.open(filename, "rt", encoding="utf-8") as f:
            support_dict = json.load(f)
        support_dict = {int(key): _fix_keys(dict) for key, dict in support_dict.items()}
        support_dicts.append(support_dict)

    filename = f"{PACKAGEDIR}/data/TESS_wcs_sip.json.bz2"
    with bz2.open(filename, "rt", encoding="utf-8") as f:
        sip = json.load(f)
    sip_dict = {
        key: _fix_keys({int(key): _fix_keys(dict) for key, dict in sip_dict.items()})
        for key, sip_dict in sip.items()
    }
    support_dicts.append(sip_dict)
    return support_dicts


def _load_warp_matrices():
    log.debug("Loading warp matrices dictionary from file")
    filename = f"{PACKAGEDIR}/data/TESS_wcs_Ms.json.bz2"
    if not os.path.isfile(filename):
        log.warn(
            "No warp matrices found. Either download them or fit them using `tesswcs.tesswcs._build_warp_matrices`"
        )
        return None, None
    with bz2.open(filename, "rt", encoding="utf-8") as f:
        Ms = json.load(f)
    Ms = {int(key): _fix_keys(dict) for key, dict in Ms.items()}
    filename = f"{PACKAGEDIR}/data/TESS_wcs_offset_weights.json.bz2"
    if not os.path.isfile(filename):
        log.warn(
            "No warp matrices found. Either download them or fit them using `tesswcs.tesswcs._build_warp_matrices`"
        )
        return None, None
    with bz2.open(filename, "rt", encoding="utf-8") as f:
        offset_weights = json.load(f)
    offset_weights = {int(key): _fix_keys(dict) for key, dict in offset_weights.items()}
    return Ms, offset_weights


def angle_to_matrix(theta):
    """Convert an angle theta in degrees to a rotation matrix"""
    cos_angle = np.cos(np.deg2rad(theta))
    sin_angle = np.sin(np.deg2rad(theta))
    return np.array([[cos_angle, -sin_angle], [sin_angle, cos_angle]])


def angle_from_matrix(matrix):
    """Convert a rotation matrix to an angle in degrees"""
    theta = np.rad2deg(np.arctan2(matrix[1, 0], matrix[0, 0]))
    return theta


def footprint(npoints=50, rows=2078, columns=2136):
    """Gets the column and row points for CCD edges"""
    column = np.hstack(
        [
            np.zeros(npoints),
            np.linspace(0, rows, npoints),
            np.linspace(0, rows, npoints),
            np.ones(npoints) * rows,
        ]
    )
    row = np.hstack(
        [
            np.linspace(0, columns, npoints),
            np.zeros(npoints),
            np.ones(npoints) * columns,
            np.linspace(0, columns, npoints),
        ]
    )
    return np.vstack([row, column]).T


def get_M(truth, approx):
    """Finds the transformation matrix that makes `approx` best fit `truth`"""
    truth = np.hstack([truth, np.ones((truth.shape[0], 1))])
    approx = np.hstack([approx, np.ones((approx.shape[0], 1))])
    a, b, e = np.linalg.solve(truth.T.dot(truth), truth.T.dot(approx[:, 0]))
    c, d, f = np.linalg.solve(truth.T.dot(truth), truth.T.dot(approx[:, 1]))

    M = np.linalg.inv(np.asarray([[a, b, e], [c, d, f], [0, 0, 1]]))
    return M


def plot_geometry(ax=None):
    # This lets us reorganize four coordinate corners to plot a square
    s = [0, 1, 3, 2, 0]

    xs, ys, xcent, ycent, M, _ = _load_support_dicts()
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(5, 5))
    for camera in np.arange(1, 5):
        for ccd in np.arange(1, 5):
            ax.plot(
                ys[camera][ccd][s],
                xs[camera][ccd][s],
                c=f"C{camera-1}",
            )
            ax.scatter(
                ys[camera][ccd].mean(),
                xs[camera][ccd].mean(),
                c=f"C{camera-1}",
            )
    lim = 60
    ax.set(
        xlim=(-lim, lim),
        ylim=(-lim, lim),
        xlabel="Column [degrees]",
        ylabel="Row [degrees]",
        title="TESS Observatory Geometry",
    )
    return ax



# Standard library
import os  # noqa

PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

from importlib.metadata import version, PackageNotFoundError  # noqa


def get_version():
    try:
        return version("tesswcs")
    except PackageNotFoundError:
        return "unknown"


__version__ = get_version()


# Standard library
import logging  # noqa: E402

log = logging.getLogger("tesswcs")

import numpy as np  # noqa: E402
from astropy.table import Table  # noqa: E402

pointings = Table.read(f"{PACKAGEDIR}/data/pointings.csv", encoding="utf-8-sig")

# Real rows and columns in a CCD
rrows, rcolumns = (2078, 2136)

# pixel size in mm
pixel_size = 0.015

# pixel scale in degrees
pixel_scale = 21 / 3600

# This sets the orientations of the ccds within the cameras
cdelt = {1: [-1, 1], 2: [-1, 1], 3: [1, -1], 4: [1, -1]}

# This sets up the corners of each CCD
pixel_corners = np.array(
    [
        [0, 0],  # Bottom left
        [0, 1],  # Top left
        [1, 0],  # Bottom right
        [1, 1],  # Top right
    ]
) * np.asarray((rrows, rcolumns))

from .tesswcs import WCS  # noqa: E402, F401
from .utils import _load_wcs_database  # noqa: E402, F401

latest_sector = np.max(list(_load_wcs_database().keys()))



"""Functions for building the database of inputs"""

import bz2
import json
import os

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.coordinates import Angle, SkyCoord
from botocore.exceptions import ClientError
from tqdm import tqdm

from . import PACKAGEDIR, log, pixel_corners, pointings, rcolumns, rrows
from .utils import _load_wcs_database, get_M


def update_database():
    _build_wcs_database()
    _build_warp_matrices()


def _build_support_dicts():
    """This function creates dictionaries of the best fitting CCD coordinates and SIP polynomials"""
    wcs_dicts = _load_wcs_database()
    # Contains the corners of each CCD w.r.t the boresight
    xs, ys = (
        {
            1: {1: [], 2: [], 3: [], 4: []},
            2: {1: [], 2: [], 3: [], 4: []},
            3: {1: [], 2: [], 3: [], 4: []},
            4: {1: [], 2: [], 3: [], 4: []},
        },
        {
            1: {1: [], 2: [], 3: [], 4: []},
            2: {1: [], 2: [], 3: [], 4: []},
            3: {1: [], 2: [], 3: [], 4: []},
            4: {1: [], 2: [], 3: [], 4: []},
        },
    )

    # Contains the center of each CCD w.r.t the boresight
    xcent, ycent = (
        {
            1: {1: [], 2: [], 3: [], 4: []},
            2: {1: [], 2: [], 3: [], 4: []},
            3: {1: [], 2: [], 3: [], 4: []},
            4: {1: [], 2: [], 3: [], 4: []},
        },
        {
            1: {1: [], 2: [], 3: [], 4: []},
            2: {1: [], 2: [], 3: [], 4: []},
            3: {1: [], 2: [], 3: [], 4: []},
            4: {1: [], 2: [], 3: [], 4: []},
        },
    )
    log.debug("Building camera/CCD data.")
    for camera in np.arange(1, 5):
        for ccd in np.arange(1, 5):
            for sector in wcs_dicts.keys():
                phi, r = (
                    wcs_dicts[sector][camera][ccd]["corner_pa"]
                    + wcs_dicts[sector]["roll"] * u.deg,
                    wcs_dicts[sector][camera][ccd]["corner_sep"].deg,
                )
                x, y = r * np.cos(phi.rad), r * np.sin(phi.rad)

                xs[camera][ccd].append(x)
                ys[camera][ccd].append(y)

                phi, r = (
                    wcs_dicts[sector][camera][ccd]["pa"]
                    + wcs_dicts[sector]["roll"] * u.deg,
                    wcs_dicts[sector][camera][ccd]["sep"].deg,
                )
                x, y = r * np.cos(phi.rad), r * np.sin(phi.rad)
                xcent[camera][ccd].append(x)
                ycent[camera][ccd].append(y)

            xs[int(camera)][int(ccd)] = list(
                np.mean(xs[camera][ccd], axis=0)
            )  # / pixel_scale
            ys[int(camera)][int(ccd)] = list(
                np.mean(ys[camera][ccd], axis=0)
            )  # / pixel_scale

            # This needs to be the crpix of the camera
            xcent[int(camera)][int(ccd)] = np.mean(xcent[camera][ccd])  # / pixel_scale
            ycent[int(camera)][int(ccd)] = np.mean(ycent[camera][ccd])  # / pixel_scale
    log.debug("Building SIP data.")
    sip = {}
    for attr in ["a", "b", "ap", "bp"]:
        sip[str(attr)] = {
            int(camera): {
                int(ccd): np.mean(
                    [
                        wcs_dicts[int(sector)][int(camera)][int(ccd)][f"sip_{attr}"]
                        for sector in wcs_dicts.keys()
                        if wcs_dicts[sector][camera][ccd]["sip_a_order"] == 4
                    ],
                    axis=0,
                ).tolist()
                for ccd in np.arange(1, 5)
            }
            for camera in np.arange(1, 5)
        }
    M = {1: {}, 2: {}, 3: {}, 4: {}}
    for camera in np.arange(1, 5):
        for ccd in np.arange(1, 5):
            truth = np.asarray(
                [
                    xs[camera][ccd] - xcent[camera][ccd],
                    ys[camera][ccd] - ycent[camera][ccd],
                ]
            )
            approx = (pixel_corners - [1045, 1001]).T
            M[int(camera)][int(ccd)] = get_M(truth.T, approx.T).tolist()

    for var in ["xs", "ys", "xcent", "ycent", "sip", "M"]:
        filename = f"{PACKAGEDIR}/data/TESS_wcs_{var}.json.bz2"
        json_data = json.dumps(locals()[var])
        with bz2.open(filename, "wt", encoding="utf-8") as f:
            f.write(json_data)
    return


def _add_sector_data(sector, c0):
    try:
        from tesscube import TESSCube
    except ImportError:
        raise ImportError(
            "Must have `tesscube` installed to build the wcs dictionary. "
            "If you installed `tesswcs` with pip make sure you include the database extras "
            "`pip install tesswcs[database]`."
        )
    sector_dict = {}
    for cam in np.arange(1, 5):
        cam = int(cam)
        sector_dict[cam] = {}
        wcss = []
        for ccd in np.arange(1, 5):
            try:
                wcss.append(TESSCube(sector, cam, ccd).wcs)
            except ClientError:
                # Data does not exist in cloud for this sector.
                return None
        for idx in range(4):
            crpix = [1045, 1001]
            crval = SkyCoord(
                *wcss[idx].wcs_pix2world([crpix], 0)[0],
                unit="deg",
            )
            sep = c0.separation(crval)
            pa = c0.position_angle(crval)
            # r = sep.deg / pixel_scale
            # phi = pa.rad
            # x, y = r * np.cos(phi), r * np.sin(phi)

            crval_sip = SkyCoord(
                *wcss[idx].all_pix2world([crpix], 0)[0],
                unit="deg",
            )
            sep_sip = c0.separation(crval_sip)
            pa_sip = c0.position_angle(crval_sip)

            sector_dict[cam][idx + 1] = {
                "pa": pa.rad,
                "sep": sep.deg,
                "pa_sip": pa_sip.rad,
                "sep_sip": sep_sip.deg,
                "ccd_center_crval": [crval.ra.deg, crval.dec.deg],
                "ccd_center_crval_sip": [crval_sip.ra.deg, crval_sip.dec.deg],
                "ccd_center_crpix": crpix,
                "crval0": list(wcss[idx].wcs.crval),
                "crpix0": list(wcss[idx].wcs.crpix),
                "cd": [list(wcss[idx].wcs.cd[jdx]) for jdx in range(2)],
                "sip_a_order": wcss[idx].sip.a_order,
                "sip_a": [
                    list(wcss[idx].sip.a[jdx])
                    for jdx in range(wcss[idx].sip.a.shape[0])
                ],
                "sip_b": [
                    list(wcss[idx].sip.b[jdx])
                    for jdx in range(wcss[idx].sip.b.shape[0])
                ],
                "sip_ap": [
                    list(wcss[idx].sip.ap[jdx])
                    for jdx in range(wcss[idx].sip.ap.shape[0])
                ],
                "sip_bp": [
                    list(wcss[idx].sip.bp[jdx])
                    for jdx in range(wcss[idx].sip.bp.shape[0])
                ],
            }
            corners = SkyCoord(
                [
                    SkyCoord(*wcss[idx].wcs_pix2world([[0, 0]], 0)[0], unit="deg"),
                    SkyCoord(
                        *wcss[idx].wcs_pix2world([[0, rcolumns]], 0)[0], unit="deg"
                    ),
                    SkyCoord(*wcss[idx].wcs_pix2world([[rrows, 0]], 0)[0], unit="deg"),
                    SkyCoord(
                        *wcss[idx].wcs_pix2world([[rrows, rcolumns]], 0)[0],
                        unit="deg",
                    ),
                ]
            )
            sector_dict[cam][idx + 1]["corner_pa"] = list(
                c0.position_angle(corners).rad
            )
            sector_dict[cam][idx + 1]["corner_sep"] = list(c0.separation(corners).deg)
            sector_dict[cam][idx + 1]["corner_ra"] = list(corners.ra.deg)
            sector_dict[cam][idx + 1]["corner_dec"] = list(corners.dec.deg)

            corners = SkyCoord(
                [
                    SkyCoord(*wcss[idx].all_pix2world([[0, 0]], 0)[0], unit="deg"),
                    SkyCoord(
                        *wcss[idx].all_pix2world([[0, rcolumns]], 0)[0], unit="deg"
                    ),
                    SkyCoord(*wcss[idx].all_pix2world([[rrows, 0]], 0)[0], unit="deg"),
                    SkyCoord(
                        *wcss[idx].all_pix2world([[rrows, rcolumns]], 0)[0],
                        unit="deg",
                    ),
                ]
            )
            sector_dict[cam][idx + 1]["corner_sip_pa"] = list(
                c0.position_angle(corners).rad
            )
            sector_dict[cam][idx + 1]["corner_sip_sep"] = list(
                c0.separation(corners).deg
            )
            sector_dict[cam][idx + 1]["corner_sip_ra"] = list(corners.ra.deg)
            sector_dict[cam][idx + 1]["corner_sip_dec"] = list(corners.dec.deg)
    return sector_dict


def _build_wcs_database():
    filename = f"{PACKAGEDIR}/data/TESS_wcs_data.json.bz2"
    if os.path.isfile(filename):
        with bz2.open(filename, "rt", encoding="utf-8") as f:
            sep_dicts = json.load(f)
    else:
        sep_dicts = {}
    for cycle, sector, ra, dec, roll, start, end in tqdm(
        pointings.iterrows(),
        desc="Loading TESS Sector",
        total=len(pointings),
        leave=True,
        position=0,
    ):
        if f"{sector}" in sep_dicts.keys():
            continue
        else:
            sector = int(sector)
            c0 = SkyCoord(ra, dec, unit="deg", frame="icrs")
            sep_dict = _add_sector_data(sector, c0)
            if sep_dict is not None:
                sep_dicts[sector] = sep_dict
                sep_dicts[sector]["ra"] = ra
                sep_dicts[sector]["dec"] = dec
                sep_dicts[sector]["roll"] = roll
            else:
                # No more sector data exists
                break
    json_data = json.dumps(sep_dicts)
    with bz2.open(filename, "wt", encoding="utf-8") as f:
        f.write(json_data)
    _build_support_dicts()


def _build_warp_matrices(plot=False):
    """Find the best fitting warp matrices

    This function uses all the archival TESS data to build a small warp matrix and updated
    CRPIX value for the WCS for each camera and CCD.

    This function works by building the true, archival WCS, and then comparing with the
    predicted WCS. We find the sky projected points in each case, and then find the transformation
    matrix that makes the predicted sky coordinates match the true sky coordinates. We then translate
    those warped coordinates back to pixel space using the predicted WCS. We can fit the warp between the
    new pixel space and the true pixel grid, to find the small change that needs to be applied to our PC matrix.

    We also find the change in CRPIX (offset term) that is best fit for each pointing.

    This term is a strong function of the roll angle w.r.t the ecliptic, so we specify a delta CRPIX as a function of the ecliptic roll.
    """
    from .tesswcs import WCS

    R, C = np.meshgrid(
        np.arange(0, rrows, 10), np.arange(0, rcolumns, 10), indexing="ij"
    )

    Ms = {
        1: {1: [], 2: [], 3: [], 4: []},
        2: {1: [], 2: [], 3: [], 4: []},
        3: {1: [], 2: [], 3: [], 4: []},
        4: {1: [], 2: [], 3: [], 4: []},
    }

    wcs_dicts = _load_wcs_database()

    for sector in tqdm(wcs_dicts.keys()):
        if plot:
            fig, ax = plt.subplots(4, 4, figsize=(10, 10), sharex=True, sharey=True)
        for camera in np.arange(1, 5):
            for ccd in np.arange(1, 5):
                wcs_t = WCS.from_sector(sector=sector, camera=camera, ccd=ccd)
                truth = wcs_t.wcs_pix2world(np.asarray([R.ravel(), C.ravel()]).T, 0)

                wcs_c = WCS.predict(
                    ra=wcs_t.ra,
                    dec=wcs_t.dec,
                    roll=wcs_t.roll,
                    camera=wcs_t.camera,
                    ccd=wcs_t.ccd,
                    warp=False,
                )
                prediction = wcs_c.wcs_pix2world(
                    np.asarray([R.ravel(), C.ravel()]).T, 0
                )

                if (wcs_t.wcs.crval[0] > 90) & (wcs_t.wcs.crval[0] < 270):
                    wrap = 360 * u.deg
                else:
                    wrap = 180 * u.deg
                truth = np.asarray(
                    [Angle(truth[:, 0] * u.deg).wrap_at(wrap).deg, truth[:, 1]]
                ).T
                prediction = np.asarray(
                    [
                        Angle(prediction[:, 0] * u.deg).wrap_at(wrap).deg,
                        prediction[:, 1],
                    ]
                ).T

                M = get_M(truth, prediction)
                transformation = M.dot(
                    np.asarray(
                        [prediction[:, 0], prediction[:, 1], prediction[:, 0] ** 0]
                    )
                )[:2].T
                pix = wcs_c.wcs_world2pix(transformation, 0)
                M2 = np.linalg.inv(get_M(np.asarray([R.ravel(), C.ravel()]).T, pix))

                wcs_c.wcs.pc = wcs_c.wcs.pc.dot(M2[:2, :2])
                prediction = wcs_c.wcs_pix2world(
                    np.asarray([R.ravel(), C.ravel()]).T, 0
                )
                prediction = np.asarray(
                    [
                        Angle(prediction[:, 0] * u.deg).wrap_at(wrap).deg,
                        prediction[:, 1],
                    ]
                ).T
                M2[:2, 2] = (
                    wcs_c.wcs_world2pix([prediction.mean(axis=0)], 0)[0]
                    - wcs_c.wcs_world2pix([truth.mean(axis=0)], 0)[0]
                )

                wcs_c.wcs.crpix += M2[:2, 2]

                prediction = wcs_c.wcs_pix2world(
                    np.asarray([R.ravel(), C.ravel()]).T, 0
                )
                prediction = np.asarray(
                    [
                        Angle(prediction[:, 0] * u.deg).wrap_at(wrap).deg,
                        prediction[:, 1],
                    ]
                ).T
                if plot:
                    sep = np.hypot(*(truth - prediction).T).reshape(R.shape)
                    ax[camera - 1][ccd - 1].pcolormesh(
                        R, C, sep, vmin=0, vmax=3 * 21 / 3600
                    )
                    ax[camera - 1][ccd - 1].set(title=f"Camera {camera}, CCD {ccd}")
                Ms[int(camera)][int(ccd)].append(M2.copy().tolist())

    ra = np.asarray([wcs_dicts[sdx]["ra"] for sdx in wcs_dicts.keys()])
    dec = np.asarray([wcs_dicts[sdx]["dec"] for sdx in wcs_dicts.keys()])
    roll = np.asarray([wcs_dicts[sdx]["roll"] for sdx in wcs_dicts.keys()])

    c = SkyCoord(ra, dec, unit="deg")
    ecl = c.transform_to("geocentricmeanecliptic")

    # The roll angle w.r.t the ecliptic plane
    ecl_roll = ecl.position_angle(
        c.directional_offset_by(-roll * u.deg, 1 * u.arcsecond).transform_to(
            "geocentricmeanecliptic"
        )
    )

    X = np.asarray([np.sin(ecl_roll), np.cos(ecl_roll), roll**0]).T

    offset_weights = {
        int(camera): {
            int(ccd): np.linalg.solve(
                X.T.dot(X), X.T.dot(np.asarray(Ms[camera][ccd])[:, :2, 2])
            ).tolist()
            for ccd in np.arange(1, 5)
        }
        for camera in np.arange(1, 5)
    }

    for camera in np.arange(1, 5):
        for ccd in np.arange(1, 5):
            Ms[int(camera)][int(ccd)] = np.median(Ms[camera][ccd], axis=0)[
                :2, :2
            ].tolist()

    filename = f"{PACKAGEDIR}/data/TESS_wcs_Ms.json.bz2"
    json_data = json.dumps(Ms)
    with bz2.open(filename, "wt", encoding="utf-8") as f:
        f.write(json_data)

    filename = f"{PACKAGEDIR}/data/TESS_wcs_offset_weights.json.bz2"
    json_data = json.dumps(offset_weights)
    with bz2.open(filename, "wt", encoding="utf-8") as f:
        f.write(json_data)
